-- Includes numerous recreations of functions that should have been exposed in the first place
-- No, seriously. Please.

local BOXTOP, BOXBOTTOM, BOXLEFT, BOXRIGHT = 1, 2, 3, 4

local function P_BoxOnLineSide(tmbox, ld)
	local p

	if ld.slopetype == "horizontal" then
		if (tmbox[BOXBOTTOM] > ld.v1.y) == (tmbox[BOXTOP] > ld.v1.y) then
			p = (tmbox[BOXTOP] > ld.v1.y)
			return p ~= (ld.dx < 0) and 1 or 0
		else
			return -1
		end
	elseif ld.slopetype == "vertical" then
		if (tmbox[BOXLEFT] < ld.v1.x) == (tmbox[BOXRIGHT] < ld.v1.x) then
			p = (tmbox[BOXRIGHT] < ld.v1.x)
			return p ~= (ld.dy < 0) and 1 or 0
		else
			return -1
		end
	elseif ld.slopetype == "positive" then
		if P_PointOnLineSide(tmbox[BOXRIGHT], tmbox[BOXBOTTOM], ld) ==
		   P_PointOnLineSide(tmbox[BOXLEFT], tmbox[BOXTOP], ld) then
			p = P_PointOnLineSide(tmbox[BOXLEFT], tmbox[BOXTOP], ld)
			return p
		else
			return -1
		end
	elseif ld.slopetype == "negative" then
		if P_PointOnLineSide(tmbox[BOXLEFT], tmbox[BOXBOTTOM], ld) ==
		   P_PointOnLineSide(tmbox[BOXRIGHT], tmbox[BOXTOP], ld) then
			p = P_PointOnLineSide(tmbox[BOXRIGHT], tmbox[BOXTOP], ld)
			return p
		else
			return -1
		end
	else
		-- default case for unexpected slopetype
		return -1
	end
end

-- Why.
local function LineBBox(ld)
	return {
		min(ld.v1.x, ld.v2.x),
		max(ld.v1.x, ld.v2.x),
		min(ld.v1.y, ld.v2.y),
		max(ld.v1.y, ld.v2.y),
	}
end

local function HighestOnLine(radius, x, y, line, slope, actuallylowest)
	-- Alright, so we're sitting on a line that contains our slope sector, and need to figure out the highest point we're touching...
	-- The solution is simple! Get the line's vertices, and pull each one in along its line until it touches the object's bounding box
	-- (assuming it isn't already inside), then test each point's slope Z and return the higher of the two.
	local v1, v2 = {}, {}
	v1.x = line.v1.x
	v1.y = line.v1.y
	v2.x = line.v2.x
	v2.y = line.v2.y

	if abs(v1.x - x) > radius then
		-- v1's x is out of range, so rein it in
		local diff = abs(v1.x - x) - radius

		if v1.x < x then -- Moving right
			v1.x = v1.x + diff
			v1.y = v1.y + FixedMul(diff, FixedDiv(line.dy, line.dx))
		else -- Moving left
			v1.x = v1.x - diff
			v1.y = v1.y - FixedMul(diff, FixedDiv(line.dy, line.dx))
		end
	end

	if abs(v1.y - y) > radius then
		-- v1's y is out of range, so rein it in
		local diff = abs(v1.y - y) - radius

		if v1.y < y then -- Moving up
			v1.y = v1.y + diff
			v1.x = v1.x + FixedMul(diff, FixedDiv(line.dx, line.dy))
		else -- Moving down
			v1.y = v1.y - diff
			v1.x = v1.x - FixedMul(diff, FixedDiv(line.dx, line.dy))
		end
	end

	if abs(v2.x - x) > radius then
		-- v2's x is out of range, so rein it in
		local diff = abs(v2.x - x) - radius

		if v2.x < x then -- Moving right
			v2.x = v2.x + diff
			v2.y = v2.y + FixedMul(diff, FixedDiv(line.dy, line.dx))
		else -- Moving left
			v2.x = v2.x - diff
			v2.y = v2.y - FixedMul(diff, FixedDiv(line.dy, line.dx))
		end
	end

	if abs(v2.y - y) > radius then
		-- v2's y is out of range, so rein it in
		local diff = abs(v2.y - y) - radius

		if v2.y < y then -- Moving up
			v2.y = v2.y + diff
			v2.x = v2.x + FixedMul(diff, FixedDiv(line.dx, line.dy))
		else -- Moving down
			v2.y = v2.y - diff
			v2.x = v2.x - FixedMul(diff, FixedDiv(line.dx, line.dy))
		end
	end

	-- Return the higher of the two points
	if actuallylowest then
		return min(
			P_GetZAt(slope, v1.x, v1.y),
			P_GetZAt(slope, v2.x, v2.y)
		)
	else
		return max(
			P_GetZAt(slope, v1.x, v1.y),
			P_GetZAt(slope, v2.x, v2.y)
		)
	end
end

local function P_MobjFloorZ(mobj, sector, boundsec, x, y, line, lowest, perfect)
	assert(mobj ~= nil)
	assert(sector ~= nil)

	if sector.f_slope then
		local testx, testy
		local slope = sector.f_slope

		-- Get the corner of the object that should be the highest on the slope
		if slope.d.x < 0 then
			testx = mobj.radius
		else
			testx = -mobj.radius
		end

		if slope.d.y < 0 then
			testy = mobj.radius
		else
			testy = -mobj.radius
		end

		if (slope.zdelta > 0) == (not lowest) then
			testx = -testx
			testy = -testy
		end

		testx = testx + x
		testy = testy + y

		-- If the highest point is in the sector, then we have it easy! Just get the Z at that point
		if R_PointInSubsector(testx, testy).sector == (boundsec or sector) then
			return P_GetZAt(slope, testx, testy)
		end

		-- If boundsec is set, we're looking for specials. In that case, iterate over every line in this sector to find the TRUE highest/lowest point
		if perfect then
			local finalheight

			if lowest then
				finalheight = INT32_MAX
			else
				finalheight = -INT32_MAX
			end

			local bbox = {
				x - mobj.radius,
				x + mobj.radius,
				y + mobj.radius,
				y - mobj.radius
			}

			for i = 0, #boundsec.lines do
				if not (boundsec.lines[i] and boundsec.lines[i].valid) then continue end
				local ld = boundsec.lines[i]
				
				local ldbbox = LineBBox(ld)
				
				if bbox[BOXRIGHT] <= ldbbox[BOXLEFT] or bbox[BOXLEFT] >= ldbbox[BOXRIGHT]
				or bbox[BOXTOP] <= ldbbox[BOXBOTTOM] or bbox[BOXBOTTOM] >= ldbbox[BOXTOP] then
					continue
				end

				if P_BoxOnLineSide(bbox, ld) ~= -1 then
					continue
				end

				if lowest then
					finalheight = min(finalheight, HighestOnLine(mobj.radius, x, y, ld, slope, true))
				else
					finalheight = max(finalheight, HighestOnLine(mobj.radius, x, y, ld, slope, false))
				end
			end

			return finalheight
		end

		-- If we're just testing for base sector location (no collision line), just go for the center's spot...
		-- It'll get fixed when we test for collision anyway, and the final result can't be lower than this
		if not line then
			return P_GetZAt(slope, x, y)
		end

		return HighestOnLine(mobj.radius, x, y, line, slope, lowest)
	else -- Well, that makes it easy. Just get the floor height
		return sector.floorheight
	end
end

local function P_MobjCeilingZ(mobj, sector, boundsec, x, y, line, lowest, perfect)
	assert(mobj ~= nil)
	assert(sector ~= nil)

	if sector.c_slope then
		local testx, testy
		local slope = sector.c_slope

		-- Get the corner of the object that should be the highest on the slope
		if slope.d.x < 0 then
			testx = mobj.radius
		else
			testx = -mobj.radius
		end

		if slope.d.y < 0 then
			testy = mobj.radius
		else
			testy = -mobj.radius
		end

		if (slope.zdelta > 0) == (not lowest) then
			testx = -testx
			testy = -testy
		end

		testx = testx + x
		testy = testy + y

		-- If the highest point is in the sector, then we have it easy! Just get the Z at that point
		if R_PointInSubsector(testx, testy).sector == (boundsec or sector) then
			return P_GetZAt(slope, testx, testy)
		end

		-- If boundsec is set, we're looking for specials. In that case, iterate over every line in this sector to find the TRUE highest/lowest point
		if perfect then
			local finalheight

			if lowest then
				finalheight = INT32_MAX
			else
				finalheight = -INT32_MAX
			end

			local bbox = {
				x - mobj.radius,
				x + mobj.radius,
				y + mobj.radius,
				y - mobj.radius
			}
			
			for i = 0, #boundsec.lines do
				if not (boundsec.lines[i] and boundsec.lines[i].valid) then continue end
				local ld = boundsec.lines[i]
				
				local ldbbox = LineBBox(ld)
				
				if bbox[BOXRIGHT] <= ldbbox[BOXLEFT] or bbox[BOXLEFT] >= ldbbox[BOXRIGHT]
				or bbox[BOXTOP] <= ldbbox[BOXBOTTOM] or bbox[BOXBOTTOM] >= ldbbox[BOXTOP] then
					continue
				end

				if P_BoxOnLineSide(bbox, ld) ~= -1 then
					continue
				end

				if lowest then
					finalheight = min(finalheight, HighestOnLine(mobj.radius, x, y, ld, slope, true))
				else
					finalheight = max(finalheight, HighestOnLine(mobj.radius, x, y, ld, slope, false))
				end

				continue
			end

			return finalheight
		end

		-- If we're just testing for base sector location (no collision line), just go for the center's spot...
		-- It'll get fixed when we test for collision anyway, and the final result can't be lower than this
		if not line then
			return P_GetZAt(slope, x, y)
		end

		return HighestOnLine(mobj.radius, x, y, line, slope, lowest)
	else -- Well, that makes it easy. Just get the ceiling height
		return sector.ceilingheight
	end
end

rawset(_G, "P_BoxOnLineSide", P_BoxOnLineSide)
rawset(_G, "HighestOnLine", HighestOnLine)
rawset(_G, "P_MobjFloorZ", P_MobjFloorZ)
rawset(_G, "P_MobjCeilingZ", P_MobjCeilingZ)